/******************************************************************************
File: drvinterface.c

******************************************************************************/

#include <signal.h>
#include <process.h>
#include <modes.h>
#include <sg_codes.h>
#include <cglob.h>  /* Definition of *_glob_data required by '_os_intercept' */
#include "i_setstt.c"
#include "i_getstt.c"
#include "status.h"    
#include "drvinterface.h"

#define DESC_NAME   "/intdescb"     /* Name of driver descriptor */

path_id pd;  /* Path for driver */
int int_read, int_write, int_end;  /* Signal handler status variables */


/******************************************************************************
SigHandler:
The signal handler that is called when the driver sends a signal to the
program.
******************************************************************************/
void SigHandler(signal_code sig_code) {
/* printf("sigcode '%x'\n", sig_code); */
  switch (sig_code)
  {
    case 2: case 3:   /* Ctrl+E or Ctrl+C signal received */
      int_end++;
      break;
    case SIG_DONEREAD:  /* Interrupt signal from driver when done reading */
      int_read++;
      break;
    case SIG_DONEWRITE:  /* Interrupt signal from driver when done writing */
      int_write++;
      break;
    default: break;
  }
  _os_rte();
}


/******************************************************************************
InitDrv: Establish connection to the driver
returns: 0 if success, 1 if connect failed
******************************************************************************/
int InitDrv() {
	/* reset states for the sighandler */
	int_read=int_write=int_end=0;
	/* assign sighandler as signal handler for this process */
	_os_intercept(SigHandler,_glob_data);
	/* connect to the driver */
	if (_os_open(DESC_NAME,FAM_READ|FAM_WRITE,&pd) != 0)
		return 1;  /* Could not open device driver */
    
    return 0;
}

/******************************************************************************
StartRead: send a message to the driver to start reading a single char to the
           address passed in pchar
returns: 0 if success in starting read, 1 if StartRead failed
******************************************************************************/
int StartRead(char *pchar) {
	int_read=int_end=0;
	/* request the device driver to read a char into pbyte,
	 * when done we will be signalled with a SIG_DONEREAD
	 */
	if (i_setstt(pd,SS_Read,(int)pchar) != 0)
		return 1;  /* Signal error */
	return 0;
}

/******************************************************************************
StartWrite: Send a message to the driver requesting the transmission of the 
            char passed in c.
returns: 0 if success in starting write, 1 if StartWrite failed
******************************************************************************/
int StartWrite(char c) {
    int_write=int_end=0;   
	/* request the device driver to write the character c, 
	 * when done we will be signalled with a SIG_DONEWRITE
	 */
    if (i_setstt(pd,SS_Write,(int)c) != 0)
	return 1;  /* Signal error */
    
    return 0;
}


/******************************************************************************
StartBufferWrite: Send a message to the driver requesting the transmission of a 
                  buffer, whose parameters are passed in pc and size.
returns: 0 if success in starting write, 1 if StartWrite failed
******************************************************************************/
int StartBufferWrite(char* pc, int size) {
	BufferCmd_t cmd;
	
	cmd.start=pc;
	cmd.end=pc+size;
	cmd.signal=SIG_DONEWRITE;
    int_write=int_end=0;   
	/* request the device driver to write the buffer cmd, 
	 * when done we will be signalled with a SIG_DONEWRITE
	 */
    if (i_setstt(pd,SS_BufWrite,(int)&cmd) != 0)
	return 1;  /* Signal error */
    
    return 0;
}


/******************************************************************************
StartBufferRead: Send a message to the driver requesting the reception of max.
                 of size bytes placing them in the memory area pointed to by
				 pc. For each byte read from the port, a signal is received.
returns: 0 if success in starting write, 1 if StartWrite failed
******************************************************************************/
int StartBufferRead(char* pc, int size) {
	BufferCmd_t cmd;
	
	cmd.start=pc;
	cmd.end=pc+size;
	cmd.signal=SIG_DONEREAD;
    int_read=int_end=0;   
	/* request the device driver to write the buffer cmd, 
	 * for each char, a SIG_DONEREAD will be received.
	 */
    if (i_setstt(pd,SS_BufRead,(int)&cmd) != 0)
	return 1;  /* Signal error */
    
    return 0;
}

/******************************************************************************
WaitRead:
Wait for a read to complete
******************************************************************************/
int WaitRead(void) {
    int status, unknown_sig;
    u_int32 ticks;
    
    do {
		unknown_sig = 0;
		_os_sigmask(1);

		if (!int_end && !int_read) {
		    ticks = 0;
		    _os9_sleep(&ticks); /* Clear signal mask and wait for signal */
		}
		
	    if (int_read) {
		    int_read = 0;  /* Interrupt signal */
		    status = 0;
		    _os_sigmask(0);
		} else if (int_end)
		    status = 1;  /* End status */
		else
		    unknown_sig = 1;  /* Unknown signal status */
			
	} while (unknown_sig);  /* Continue until a known signal is received */
    return status;
}

/******************************************************************************
WaitReads: Waits for any number of reads to complete...

******************************************************************************/
int WaitReads(int nofreads) {
    int status, unknown_sig;
    u_int32 ticks;
    
    do {
		unknown_sig = 0;
		_os_sigmask(1);

		if (!int_end && !int_read) {
		    ticks = 0;
		    _os9_sleep(&ticks); /* Clear signal mask and wait for signal */
		}
		
	    if (int_read) {
		    int_read = 0;  /* Interrupt signal */
		    status = 0;
		    _os_sigmask(0);
		} else if (int_end)
		    status = 1;  /* End status */
		else
		    unknown_sig = 1;  /* Unknown signal status */
			
	} while (unknown_sig);  /* Continue until a known signal is received */
    return status;
}


/******************************************************************************
WaitWrite:
Waits for a known signal from the signal handler.
******************************************************************************/
int WaitWrite(void) {
	int status, unknown_sig;
	u_int32 ticks;

  do {
    unknown_sig = 0;
    _os_sigmask(1);
    if (!int_end && !int_write) {
      ticks = 0;
      _os9_sleep(&ticks); /* Clear signal mask and wait for signal */
    }
    if (int_write) {
      int_write = 0;  /* Interrupt signal */
      status = 0;
      _os_sigmask(0);
    }
    else
      if (int_end)
        status = 1;  /* End status */
      else
        unknown_sig = 1;  /* Unknown signal status */
  }
  while (unknown_sig);  /* Continue until a known signal is received */
  return status;
}

/******************************************************************************
CloseDrv: Sends a message to the driver to cancel our pending reads (if any) and 
          disconnects us from the driver.
******************************************************************************/
int CloseDrv(void) {
	int status;

	i_setstt(pd,SS_Cancel_Read,0);
	i_setstt(pd,SS_Cancel_Write,0);
	status = _os_close(pd);
	return status;
}
