/*****************************************************************************
File: intdrvb.c

Exercise 5: Interrupt driver.
*****************************************************************************/

#include <errno.h>      /* error codes */
#define E_UNKSVC	0xd0	/* Unknown Service Code */
#define E_NOTRDY	0xf6	/* Device Not Ready */

#include <MACHINE/reg.h>  /* register stack frame */
#include <procid.h>       /* process descriptor */
#include <sg_codes.h>     /* get/set status codes */
#include <path.h>         /* path descriptor structure */
#include <const.h>        /* constants (e.g. NULL) */ 
#include <signal.h>       /* to handle signals */
#include <scf.h>    /* definition for SCF static storage, MUST be present */
#include "status.h"    

extern int IRQSvc(void *);    /* function in 'dskel.a' */

struct scfstatic scfs;    /* SCF static storage declaration, MUST be
                             present, MUST be declared first */
void *sysglobs;        /* used by 'dskel.a' to store pointer to sysglobs */
Pathdesc pathdesc;     /* path descriptor ptr */
procid *procdesc;      /* used by 'dskel.a' to store ptr to process desc */
int errno;             /* holds error number */

/* Mem I/O of serial port 2 (B) */
u_int8 *portBreg; /* normal access - reg0 derefter register */
u_int8 *portBio;  /* direkte til register 8 */

/* destination for read request */
process_id proc_target;
char* target_pos;
BufferCmd_t target;

/* buffered write source */
process_id proc_source;
BufferCmd_t source;

void (*intsrv)(void);
int f_irq(unsigned char,unsigned char,int (*)(void *),char *);

/*****************************************************************************
Function name: Init

Purpose: Initialize timer, set up interrupt

Input:  device descriptor pointer

Output: error code

Side effects: Set pointers to timer registers.
        Change timer registers.
        Insert ISR into vector table.

History: Created Jan 20, 1993 by CA 
*****************************************************************************/
int init(mod_dev *dd) {
	portBreg = (void*)0xfffe3000;
	portBio = (void*)0xfffe3001;
	
	proc_source=0;
	source.start=source.end=NULL;
	source.signal=SIG_DONEWRITE;
	
	proc_target=0;
	target.start=target.end=NULL;
	target.signal=SIG_DONEREAD;
	
	/* install the interrupt service routine */
	if( (f_irq(dd->_mvector, dd->_mpriority, IRQSvc, dd->_mport)) == ERROR )
		return(errno);

	/* Initialize portB */

	/* Reset in WR9 (Master Int. Control)
	 * ChannelB reset,  [7:6]=01
	 * Software int. ack, [5]=0 reserved on NMOS
	 * Status High/Low, [4]= 0
	 * Resten = 0 (naar int. er slaaet fra)
	 */
	*portBreg=0x09; *portBreg=0x40;
	
	
	/* set WR4 (transmit/receive modes) to:
	 * clock mode - 16x, [7:6]=01
	 * sync char - 8-bit, [5:4]=00  *ignored i asyncmode
	 * stop bits - 1, [3:2]=01
	 * parity - disabled [1:0]=00
	 */
	*portBreg=0x04; *portBreg=0x44;
	
	/* set WR3 (receive params) to:
	 * bits per char - 8, [7:6]=11
	 * auto enable - disabled, [5]=0
	 * hunt mode - disabled, [4]=0   *ignored i asyncmode
	 * receive CRC - disabled, [3]=0   *ignored i asyncmode
	 * address search mode - disabled, [2]=0   *ignored i asyncmode
	 * sync char load inhibit - disabled, [1]=0
	 * receiver enabled - disabled, [0]=0
	 */
	*portBreg=0x03; *portBreg=0xC0;
	 
	/* set WR5 (transmit params) to:
	 * DTR control - enabled, [7]=1
	 * char select bits - 8, [6:5]=11
	 * send break - disabled, [4]=0
	 * transmit enabled - disabled, [3]=0 * while initing
	 * SDLC/CRC16 - unset, [2]=0  *ignored i asyncmode
	 * RTS control - disabled, [1]=0
	 * transmit CRC enable - disabled, [0]=0  *ignored i asyncmode
	 */
	*portBreg=0x05; *portBreg=0xE0;
	
	/* set WR1 (interrupt and data transfer):
	 * wait/dma request enable - , [7]=0
	 * wait/dma request func. - , [6]=0
	 * wait/dma request on recv/trans - , [5]=0
	 * Rx int. - on all chars and spec. cond., [4:3]=10
	 * parity is special condition - no, [2]=0
	 * Tx int enable - yes, [1]=1
	 * ext. int. enable - no, [0]=0
	 */
	*portBreg=0x01; *portBreg=0x12;
	 
	/* set WR2 (interrupt vector)
	 * vector -  88_dec, [7:0]=01011000
	 */
	*portBreg=0x02; *portBreg=0x58;
	 
	/* set WR10 (recv./trans. control bits)=0
	 * CRC preset - [7]=0  *ignored i asyncmode
	 * data enc. - NRZ, [6:5]=00
	 * go-active-on-poll - 0, [4]=0  *ignored i asyncmode
	 * mark flag idle ctrl - 0, [3]=0   *ignored i asyncmode
	 * abort/flag on underrun - 0, [2]=0   *ignored i asyncmode
	 * loop mode - disabled, [1]=0   *ignored i asyncmode
	 * sync select, [0]=0   *ignored i asyncmode
	 */
	 *portBreg=0x0A; *portBreg=0x00;
	 
	 /* set WR11 (clock mode control)
	  * BR generator (not nescessary)
	  */
/*	 *portBreg=0x0B; *portBreg=0x; */
	  
	/* set WR12,WR13 (baud rate control)
	 * time constant = 14
	 * WR12=14
	 * WR13=0
	 */
	 *portBreg=0x0D; *portBreg=0x00; 
	 *portBreg=0x0C; *portBreg=0x0E;
	
	/* set WR14 (misc. control bits
	 * dig. phase-locked loop ctrl - null command, [7:5]=000
	 * local loop-back enable - no, [4]=0
	 * auto echo - no, [3]=0
	 * DTR... - [2]=0
	 * BR generator source - , [1]=1
	 * BR generator enable - yes, [0]=1
	 */
	*portBreg=0x0E; *portBreg=0x03;
	
	/* WR15 irrelevant */
	
	/* set WR9 (master interrupt control)
	 * ChannelB reset,  [7:6]=00
	 * Software int. ack, [5]=0 reserved on NMOS
	 * Status High/Low - low, [4]= 0
	 * Master Interrupt Enable - set, [3]=1
	 * Disable lower chain ctrl - unset, [2]=0
	 * No vector select - unset[1]=0
	 * Vector Includes Status ctrl - unset[0]=0
	 */
	*portBreg=0x09; *portBreg=0x08;

	/* enable transmitter in WR5 */
	*portBreg=0x05; *portBreg=0xE8;
		
	/* enable receiver in WR3*/
	*portBreg=0x03; *portBreg=0xC1;

	return 0;
}


/*****************************************************************************
Function name: read
 
Purpose: Read a character from device, not implemented, unused

Input: none

Output: errorcode

Side effects:

History: Craeted Jan20, 1993 by CA
*****************************************************************************/
int read(void) {
	return -1;
}


/*****************************************************************************
Function name: write

Purpose: write a character to device, not implemented, currently unused

Input:

Output:

Side effects:

History: Created Jan 20, 1993 by CA
*****************************************************************************/
int write(char c) {
	return 0;
}


/*****************************************************************************
Function name: getstat

Purpose: replies to status requests

Input: function to perform, callers stack frame
       The REGISTERS *r layout:
         r->d[0] callers 1. parameter (the path)
         r->d[1] callers 2. parameter (the SS_ code)
         r->d[2] callers 3. parameter (first parameter to be passed)
         r->d[3] callers 4. parameter 
         r->d[4] callers 5. parameter 

Output: reply; result of function to perform

Side effects: none

History: Created Jan 20, 1993 by CA
*****************************************************************************/
int getstat(int func, REGISTERS *r) {
	errno = 0;
	switch(func) {

		default: 
		    errno = E_UNKSVC;
		    break;
	}
    return errno;
}



/*****************************************************************************
Function name: setstat

Purpose: set device status

Input: function to perform, callers stack frame
       The REGISTERS *r layout:
         r->d[0] callers 1. parameter (the path)
         r->d[1] callers 2. parameter (the SS_ code)
         r->d[2] callers 3. parameter (first parameter to be passed)
         r->d[3] callers 4. parameter 
         r->d[4] callers 5. parameter 

Output: error code

Side effects: may change static variables and device hardware registers

History: Created Jan 20, 1993 by CA
*****************************************************************************/
int setstat(int func, REGISTERS *r) {
	char c;
	
	errno = 0;
	switch(func) {
		/* user requests read from device - target address passed */
		case SS_Read: 
			/* check that nobody else has a pending read (is waiting for a signal) */
			if (proc_target == 0) { 
				/* the target to store the character read is passed in r->d[2] */				
			    target.start = (char*) r->d[2];
				target.end = target.start+1;
				target_pos=target.start;
				target.signal = SIG_DONEREAD;
			    /* get the process id to signal */
			    proc_target = procdesc->_id;
			} else {
				/* somebody else is waiting for a signal, return error */
				errno = E_NOTRDY;
			}
			break;
			
		/* user requests a buffered read from the device */
		case SS_BufRead:
			/* make sure nobody else is using the device */
			if ( (proc_target==0) || (proc_target==procdesc->_id) ) {
				proc_target=0;
				target = *((BufferCmd_t*)r->d[2]);
				target_pos=target.start;
				proc_target=procdesc->_id;
			} else {
				errno = E_NOTRDY;
			}
			break;
		
		/* users requests a write to the device - char passed */
		case SS_Write: 
			/* check that nobody else already requested a write */
			if( proc_source == 0 ) {
			    c = (char) r->d[2]; /* get the char to send */
			    /* get the process id to signal */
			    proc_source = procdesc->_id;
				/* make sure a buffered write is not attempted */
				source.start=NULL;
				source.end=NULL;
				source.signal=SIG_DONEWRITE;
				/* put c in FIFO */
				*portBio=c;
			} else {
			    /* somebody else has a pending write and is waiting for a signal, return error */
			    errno = E_NOTRDY;
			}
		    break;
		
		
		/* Start a buffered write */
		case SS_BufWrite:
			if( proc_source==0 ) {
				/* get the buffercmd from the param */
			    source = *((BufferCmd_t*)r->d[2]); 
			    /* get the process id to signal */
			    proc_source = procdesc->_id;
				/* start transmitting by putting first in FIFO */
				if (source.start<source.end) {
					*portBio=*source.start;
					source.start++;
				}
			} else {
			    /* somebody else has a pending write and is waiting for a signal, return error */
			    errno = E_NOTRDY;
			}
		    break;

		/* Do not signal this process anyway and release pending signal */
		case SS_Cancel_Read: 
		    if (proc_target==procdesc->_id)
				proc_target = 0;
			else errno=E_NOTRDY;
		    break;
			
		case SS_Cancel_Write: /* clear pending signal condition */
		    if (proc_source==procdesc->_id) {
				proc_source = 0;
				source.start=source.end=NULL;
			} else 
				errno=E_NOTRDY;
		    break;
			
			

		/* unknown or not implemented function code, return error */	    
		default:
			errno=E_UNKSVC;
			break;
	}
    return errno;
}


/*****************************************************************************
Function name: term

Purpose: de-initialize device

Input: device descriptor pointer

Output: error code

Side effects: Remove ISR from vector table.
        resets device registers.

History: Created Jan 20, 1993 by CA
*****************************************************************************/
int term(mod_dev *dd) {
	/* Disable portB interrupts */
	*portBreg=9; *portBreg=0x00;
	/* Disable portB transmitter and receiver */
	*portBreg=3; *portBreg=0xE8; /*11101000*/
	*portBreg=5; *portBreg=0xC0; /*11000000*/

	/* remove interrupt handler */
	f_irq(dd->_mvector,0,NULL,NULL);
	return 0;
}


/*****************************************************************************
Function name: irqsvc

Purpose: interrupt service routine, handles interrupt status clear and
     signal eventual waiting process.
	 This function is only called upon a int. 88!

Input: base address of device

Output: success if our device was the interrupting device, otherwise failure

Side effects: clear interrupt status.
        If a process is waiting for a signal, the signal is send
          and the process ID is cleared from the table.

History: Craeted Jan 20, 1993 by CA
*****************************************************************************/
int irqsvc(void *port) {
    u_char status;
	u_int8 *portBreg_ = (void*)0xfffe3000; /* normal access - reg0 derefter register */
	u_int8 *portBio_ = (void*)0xfffe3001; /* direkte til register 8 */
	char c;

    /* get cause of interrupt (status of our device and the reason for the interrupt) */
	*portBreg_=2;
	status = *portBreg_ & 0x0E; /* pick our bits only */

/*	_os_send(proc_write,0x8000+status);*/
	switch (status) {
		/* Transmit buffer empty */
		case 0:
			if (source.start<source.end) {
				*portBio=*source.start;
				source.start++;
			} else {
				/* clear the pending Tx-buffer-empty interrupt */
				*portBreg=0x28; /*00101000*/
				/* done doing the write, so send signal to write-process of completed write */
			    if( proc_source != 0 ) {
				    _os_send(proc_source,source.signal);
				    proc_source = 0;
				}
			}
			break;
		
		/* External status change */
		case 2:
			break;
		
		/* Receive char ready */
		case 4:
			c=*portBio;
		    if( proc_target != 0 ) {
				if (target_pos<target.end) {
					*(target_pos)=c;
					target_pos++;
					if (target_pos>=target.end) target_pos=target.start;
				}
			    _os_send(proc_target,target.signal);
			}
			break;
		
		/* special receive condition */
		case 6:
			break;
		
		default:
			return -1;
	}

	/* signal that interrupt has been handled */
	/* Reset Highest IUS: Set WR0[5:3]=111, WR0=0b00111000*/
	*portBreg_=0x38; 
	
    return 0;
}




/*****************************************************************************
Function name: f_irq

Purpose: hook to F$IRQ

Input: interrupt vector, polling priority, pointer to ISR, device base address

Output: error code

Side effects: installs ISR in OS9 interrupt table

History: Created Jan 20, 1993 by CA
*****************************************************************************/
_asm("f_irq:");
_asm("  movem.l d1/a0/a2-a3,-(a7)"); /* save registers, d0 and d1 already */
                   /* contains correct parameters */
_asm("  movem.l 20(a7),a0/a3");     /* get handler and port address */
_asm("  movea.l a6,a2");        /*  copy static storage address */
_asm("  OS9 F$IRQ");         /*  install ISR */
_asm("  bcs.s  f_irq10");        /*  ..error */
_asm("  moveq #0,d0");          /*  no error, return 0 */
_asm("  bra.s  f_irq20");
_asm("f_irq10");
_asm("  move.l  d1,errno(a6)");     /*  save OS-9 error code */
_asm("  moveq #-1,d0");         /*  indicate error */
_asm("f_irq20");
_asm("  movem.l  (a7)+,d1/a0/a2-a3"); /* retrieve registers */
_asm("  rts");


