#ifndef LAYER1_H
#define LAYER1_H

#include "types.h"

#define PREAMBLE '\x1e'
#define FORWARD_BUFFER_SIZE 16
#define IN_BUFFER_SIZE 8
#define OUT_BUFFER_SIZE 8
#define OUT_STATE_FREE 0
#define OUT_STATE_NEW 1
#define OUT_STATE_TRANS 2
#define OUT_STATE_ACK 3
#define OUT_STATE_NAK 4

/* Native network format of frames transmitted */
typedef struct {
	u_char preamble;          /* signatur for start-af-frame 0x1e */
	unsigned dest: 4;       /* destination for frame */
	unsigned source: 4;     /* source for frame !=0 */
	unsigned ack: 1;        /* acknowledge (receiver got frame intact) */
	unsigned sn: 3;         /* currently unused - serial (mod 8) number of */
	                        /* packet (max 8 outstanding frames)) */
	unsigned TTL: 4;        /* Time-To-Live (decremented in each forwarding. */
	                        /* Frame discarded if TTL=0 */
	unsigned size: 8;       /* Size of data following this frame */
	char data[255];         /* ...data... */
	u_char checksum;          /* some sort of checksum created for this frame */	                        /* (header plus data) */
} frame_t;


extern frame_t in_buffer[IN_BUFFER_SIZE];
extern int in_start; /* points to most recently received frame */
extern int in_end; /* points to oldest frame in queue */
extern int in_changed; /* counts the number of changes made to in_buffer */

/* There are 8 entries in the out-buffer, which are always used in
 * order. Therefore, if entry 2,3 is transmitted and ACK/NAK arrived on
 * 3, 2 must be a NAK (since the order of frames is never changed).
 */
extern frame_t out_buffer[OUT_BUFFER_SIZE];
/* The state of each out-buffer is recorded in this array. The states are> */
extern char out_state[OUT_BUFFER_SIZE]; /* 0 = free, 1=new, 2=transmitted, 3=acked, 4=timeout/failed */
extern int out_transmitted; /* first un-transmitted frame, look forward for free buffers, at most till first transmitted! */
extern int out_changed; /* number of frames changed due to received ack/nak */

extern int forward_changed;


/* Remember to call this method with the network ID before starting recv.
 * param:  m_addr - address on network
 * return: E_OK - successfully inited
 */
int l1_init( int m_addr );


/* Add a frame to the frame out buffer.
 * The frame is assigned a number in the buffer, that can be polled for status.
 * param : f - pointer to frame to send
 * return: E_NOFREEBUFFERS - if all out-buffers are in use
 *         [frame n] >0 - if assigned a frame
 */
int l1_sendFrame(frame_t *f);


/* Frees a frame in the buffer when it is ACK or NAK. 
 * param : no - frame number to free
 * return: E_OK - frame freed
 *         E_INVALIDFRAMESTATE - entry is not ACK or NAK
 */
int l1_freeFrame(int no);


/* Gets the status of the buffer. See OUT_STATE_??? for states
 * param : no - number of buffer to poll
 * return: state from OUT_STATE_???
 */
int l1_getFrameStatus(int no);


/* Force a reset output buffer (never do this unless you are sure the net is empty) 
 * param :
 * return:
 */
void l1_flushBuffer(void);

/* Get index of a free location in the out_buffer to store a new frame.
 * param : f - pointer to location to store index of free entry in out_buffer
 * return: E_OK - a free frame was found
 *         E_NOFRAMESAVAIL - no free frames were found in the out-buffer
 */
int l1_getFreeOutFrame(int *f);


/* If any frames are present in the in-buffer, a pointer to the last frame is 
 * written to the location pointed to by f.
 * param : f - pointer to where to store pointer to frame
 * return: E_OK - if successfully copied frame
 *         E_NOFRAMESAVAIL - if the in-buffer is empty
 */
int l1_peekInFrame(frame_t* *f);


/* If any frames are present in the in-buffer, the last frame is dequeued.
 * return: E_OK - if successfully copied frame
 *         E_NOFRAMESAVAIL - if the in-buffer is empty
 */
int l1_doneInFrame(void);


/* If any forward-buffers are present, they are prepared in the buffer passed and
 * the length of network data is returned. 
 * param : buffer - location of buffer (expected to be large enough <550)
 * return: 0 - no forward pending
 *         length of data outputted,
 */
int l1_prepareForward(char* buffer);

/* If any frames are ready in the out_buffer (and no frames are waiting to
 * be forwarded), data is prepared for the network in the buffer supplied.
 * The length of network data is returned. 
 * param : buffer - location of buffer (expected to be large enough <550)
 * return: 0 - no outs are pending (or forwards pending)
 *         length of data outputted,
 */
int l1_prepareOutBuffer(char* buffer);

/* Call onReadChar for each byte received from the network. onReadChar will 
 * perform desynching and correctly detect start-of-frames. When onReadChar
 * completes a frame and accept it as OK, forward_changed will be incremented.
 * Use forward_changed in your client side to detect new frames for forwarding (or
 * use prepareForward() if you rather do that).
 * param : c - char received from network.
 * return: ...
 */
void l1_onReadChar(char c);


#endif
