/* File: layer1.c
 * This module handles network traffic and forwarding. 
 * Frames for us are put in the in-buffer (if room) and ACKed on the network.
 * Frames from us (to transmit) are put in the out-buffer and transmitted when
 * the network has room for it.
 */
#include "layer1.h"

/****************************************************************************
  B U F F E R S
  These globals declare the buffers used for forwarding, received and 
  transmitted buffers.
****************************************************************************/

/* The forward buffer: */
frame_t forward_buffer[FORWARD_BUFFER_SIZE];
int forward_outpos; /* currently transmitting this frame */
int forward_inpos; /* buffer being filled by network driver */

/* see header */
frame_t in_buffer[IN_BUFFER_SIZE];
int in_start; /* points to most recently received frame */
int in_end; /* points to oldest frame in queue */

frame_t out_buffer[OUT_BUFFER_SIZE];
char out_state[OUT_BUFFER_SIZE]; /* 0 = free, 1=new, 2=transmitted, 3=acked, 4=timeout/failed */
int out_transmitted; /* last transmitted frame, look forward for free buffers, at most till first transmitted! */


/****************************************************************************
  B U F F E R   M A N A G E M E N T
  The functions in this section is concerned with handling the in- and out-
  buffers, and should be used when enqueing further frames.
****************************************************************************/


/* Add a frame to the frame out buffer.
 * The frame is assigned a number in the buffer, that can be polled for status.
 * param : f - pointer to frame to send
 * return: E_NOFREEBUFFERS - if all out-buffers are in use
 *         [frame n] >0 - if assigned a frame
 */
int sendFrame(frame_t *f) {
	int freefr=out_transmitted;
	
	do {
		if (out_state[freefr]==OUT_STATE_FREE) {
			out_state[freefr]=OUT_STATE_NEW;
			out_buffer[freefr]=*f;
			return freefr;
		}
		freefr++;
		if (freefr>=OUT_BUFFER_SIZE) freefr=0; /* wraparound */
	} while (freefr!=out_transmitted);
	
	return E_NOFREEBUFFERS; /* no available entry */
}


/* Frees a frame in the buffer when it is ACK or NAK. 
 * param: no - frame number to free
 * return: E_OK - frame freed
 *         E_INVALIDFRAMESTATE - entry is not ACK or NAK
 */
int freeFrame(int no) {
	if (out_state[no]==OUT_STATE_ACK || out_state[no]==OUT_STATE_NAK) {
		out_state[no]=OUT_STATE_FREE;
		return E_OK;
	} else
		return E_INVALIDFRAMESTATE;
}


/* Gets the status of the buffer. See OUT_STATE_??? for states
 * param : no - number of buffer to poll
 * return: state from OUT_STATE_???
 */
int getFrameStatus(no) {
	return out_state[no];
}


/* Force a reset output buffer (never do this unless you are sure the net is empty) 
 * param :
 * return:
 */
void flushBuffer() {
	int lp1;
	for (lp1=0;lp1<OUT_BUFFER_SIZE;lp1++)
		out_state[lp1]=OUT_STATE_FREE;
}


/* Checks for new frames in in-buffer.
 * param :
 * return: 1 if frames are present
 *         0 if no frames in buffer
 */
int hasInFrames() {
	return !(in_end==in_start);
}


/* If any frames are present in the in-buffer, the last frame is dequeued and
 * copied to the location provided.
 * param : f - pointer to where to store frame
 * return: E_OK - if successfully copied frame
 *         E_NOFRAMESAVAIL - if the in-buffer is empty
 */
int getFrame(frame_t *f) {
	if (in_start==in_end) 
		return E_NOFRAMESAVAIL;
	else {
		*f = in_buffer[in_end];
		in_end++;
		if (in_end>=IN_BUFFER_SIZE) in_end=0; /* wraparound */
		return E_OK;
	}
}

/* Put a frame in the in-buffer
 * param : f - pointer to frame to add to queue.
 * return: E_OK - if success
 *         E_NOFREEBUFFERS - if in-buffer is overloaded
 */
int putInFrame(frame_t *f) {
	if ( (in_start+1)%IN_BUFFER_SIZE==in_end ) /* out of buffers */
		return E_NOFREEBUFFERS;
	else {
		in_buffer[in_start] = *f;
		in_start++;
		if (in_start>=IN_BUFFER_SIZE) /* wrap-around */
			in_start=0;
		return E_OK;
	}
}



/****************************************************************************
  L A Y E R - 0   F U N C T I O N S
  These are the functions related to lowest level handling of the driver and
  the network format of the data.
****************************************************************************/

/* This function will format the frame to the network format in a buffer 
 * ready to send (perform de-sync and checksum calculation).
 * It is assumed, that enough room are available in out_buffer.
 * param : f - frame to convert
 *         out - location of buffer (expected to be large enough <550)
 * return: length of data outputted
 */
int formatFrame(frame_t *f, char *out) {
	char* fp;
	int inlen = f->size+3; /* the three bytes from the header */
	int outlen = 0;
	char c;
	u_char checksum = 0;
	
	fp = (char*) f;
	
	/* transfer the preamble without desynching */
	*(out++) = *(fp++); outlen++;
	for (;inlen>0;inlen--) {
		/* transfer byte and do desynching */
		c = *(fp++);
		*(out++) = c; outlen++;
		/* update the checksum */
		checksum+=c;
		if (c==PREAMBLE) {
			*(out++) = 0; outlen++;
		}
	}
	/* transmit the checksum and perform de-syncing if nescessary */
	*(out++) = checksum; outlen++;
	if (checksum==PREAMBLE) {
		*out = 0; outlen++;
	}
	/* update the checksum field of the frame */
	f->checksum=checksum;
	
	return outlen;
}

/* If any forward-buffers are present, they are prepared in the buffer passed and
 * the length of network data is returned. 
 * param : buffer - location of buffer (expected to be large enough <550)
 * return: 0 - no forward pending
 *         length of data outputted,
 */
int prepareForward(char* buffer) {
	int len;
	
	if (forward_inpos!=forward_outpos) {
		/* format frame and start write */
		len=formatFrame(&forward_buffer[forward_outpos], buffer);
		forward_outpos++;
		if (forward_outpos>=FORWARD_BUFFER_SIZE) /* wrap-around */
			forward_outpos=0;
		return len;
	} else 
		return 0;
}

/* marker that the frame is done */
int frameDone;
/* my network address */
unsigned int my_addr;
/* points to the frame currently being filled with data */
static frame_t *fw_curr;
/* points into the frame currently being filled (a byte-based pointer for 
 * higher performance */
static char *fw_bufp;
/* last character received */
static char last_char=0;
/* checksum so far */
static u_char checksum=0;
/* number of chars received in this frame */
static int length=0;

#define DEBUG
#ifdef DEBUG
#include <stdio.h>
#endif

void onFrame(void);

/* Call onReadChar for each byte received from the network. onReadChar will 
 * perform desynching and correctly detect start-of-frames. When onReadChar
 * completes a frame and accept it as OK, frameDone will be incremented.
 * Use frameDone in your client side to detect new frames for forwarding (or
 * use prepareForward() if you rather do that).
 * param : c - char received from network.
 * return: ...
 */
void onReadChar(char c) {
	 
	if (c==PREAMBLE) 
		last_char=c;
	else {
		if (last_char==PREAMBLE) {
			if ( c==0 ) { /* see if sender is 0 - invalid => desync */
				if ( length<sizeof(frame_t) ) { /* ensure that no frames too long can mess up */
					*(fw_bufp++) = last_char;
					length++;
					checksum+=PREAMBLE;
				}
				last_char=~PREAMBLE;
			} else { /* start of new frame */
#ifdef DEBUG
				printf("desync'er caught new sync at %d in curr. frame\n", fw_bufp-(char*)fw_curr);
#endif
				fw_curr->checksum=*(fw_bufp-1); /*copy last recv character into frame.checksum */
				checksum -= fw_curr->checksum; /* correct checksum (real checksum included) */
				/* check (and accept) frame */
				if (length) {
#ifdef DEBUG
					printf("desync: ");
#endif
					onFrame(); /* sets forward_inpos */
				}
				/* start from forward_buffer's current entry */
				fw_curr = &forward_buffer[forward_inpos];
				fw_bufp = (char*) fw_curr;
				/* put missing data in buffer */
				*(fw_bufp++) = last_char;
				*(fw_bufp++) = c;
				length = 2;
				checksum = c;
				last_char = 0;
			}
		} else { /* normal data */
			if ( length<((int)fw_curr->size+4) ) { /* end-of-header/data not reached */
				*(fw_bufp++) = c; /* add char to frame and move pointer */
				checksum += c; /* update checksum */
				length++;
			} else { /* end-of-data reached */
				fw_curr->checksum=c;
				/* check (and accept) frame */
				if (length) {
#ifdef DEBUG
					printf("length-term: ");
#endif
					onFrame();
				}
				/* set pointers ready for next frame */
				fw_curr = &forward_buffer[forward_inpos];
				fw_curr->size = 0;
				fw_curr->TTL = 0; /* make sure the desync'er discards this frame*/
				fw_bufp = (char*) fw_curr;
				length = 0;
			}
		}
	}
	
	return;
}


/* When a frame is received, this function is called, which is responsible for
 * updating forward_inpos correctly. This means, move forward_inpos forward if
 * this frame needs forwarding. It updates the frame correspondingly.
 * 
 */
void onFrame() {
	if ( PREAMBLE!=fw_curr->preamble || /* check for pre-amble correct */
	     length!=((int)fw_curr->size+4) || /* 4 extra bytes of header and checksum */
	     checksum!=fw_curr->checksum || /* checksum matches */
	     !fw_curr->TTL || /* frame still alive */
	     !fw_curr->source || /* the sender exists */
	     (fw_curr->dest==my_addr && fw_curr->ack) || /* frame must be discarded */
	     fw_curr->source==my_addr /* frame from me being sent around */ )
	{
#ifdef DEBUG
		printf("onFrame discarded frame because: \n");
		if ( PREAMBLE!=fw_curr->preamble )
			printf("\tpreamble was incorrect\n");
		if ( length!=((int)fw_curr->size+4) )
			printf("\tframe length mismath\n");
		if ( checksum!=fw_curr->checksum)
			printf("\tchecksum error\n");
		if ( !fw_curr->TTL )
			printf("\tTTL=0, frame outdated and discarded\n");
		if ( !fw_curr->source )
			printf("\tframe has no source\n");
		if ( fw_curr->dest==my_addr && fw_curr->ack )
			printf("\tpart of a round-trip\n");
	   if ( fw_curr->source==my_addr )
			printf("\tframe sent from me, ACK/NAK in out_state[] updated accordingly\n");
#endif
	   if ( fw_curr->source==my_addr )
			if ( fw_curr->ack )
				out_state[fw_curr->sn]=OUT_STATE_ACK;
			else
				out_state[fw_curr->sn]=OUT_STATE_NAK;
	} else { 
		/* drop frame when forward_buffer is overloaded */
		if ( (forward_inpos+1)%FORWARD_BUFFER_SIZE==forward_outpos ) {
#ifdef DEBUG
			printf("onFrame: out of forward buffers...\n");
#endif
			return;
		} else {
			forward_inpos++;
			frameDone++;
		}
		if (forward_inpos>=FORWARD_BUFFER_SIZE) /* wrap-around */
			forward_inpos=0;
		fw_curr->TTL--; /* reduce TTL */
		if ( fw_curr->dest==my_addr ) {
			/* frame is for me, try to put in in-buffer */
			if ( !putInFrame(fw_curr) ) {
				fw_curr->ack=1;
				fw_curr->size=0;
				fw_curr->TTL=15;
			} else {
#ifdef DEBUG
				printf("onFrame: out of in buffers...\n");
#endif
				fw_curr->ack=0;
				fw_curr->size=0;
				fw_curr->TTL=15;
			}
		}
	}
}



/****************************************************************************
  I N I T I L A Z A T I O N
  Starts the networking layer-1...
****************************************************************************/
int init( int m_addr ) {
	int lp1;
	
	forward_inpos = 0;
	forward_outpos = 0;
	frameDone = 0;
	in_start = 0; 
	in_end = 0; 

	for (lp1=0;lp1<OUT_BUFFER_SIZE;lp1++) 
		out_state[lp1]=OUT_STATE_FREE;
	out_transmitted = 0;

	fw_curr = &forward_buffer[forward_inpos];
	fw_bufp = (char*) fw_curr;
	
	my_addr = m_addr;

	return E_OK;
}
	

/*	if char is preamble	
			last char = current
			
		if last char is preamble 
			if current&0x0f is zero  (preamble in data - source is never 0!)
				put last char in frame_buffer
				last char = 0
				len++, update checksum (cs+=data)
			else (new frame)
				set flag for frame done 
				move checksum and verify frame
				if checksum or len mismatch error
					discard
				start new frame
				len=1
		else (normal data)
			move current char from driver_in_buffer to current frame in forward buffer
			len++, update checksum (cs+=data)
*/
