/* File: layer2.c
 *
 * Handles all retransmission of lost packets, signaling to processes when 
 * status of packets change
 */
 
#include <stdio.h>
#include <string.h>

#include <types.h>
#include <process.h>
#include <sg_codes.h>
#include <modes.h>
#include <signal.h>
#include <module.h>
#include <semaphore.h>
#include <time.h>
#include <cglob.h>  /* Definition of *_glob_data required by '_os_intercept' */
#include <stdlib.h> /* for atoi() */

#include "layer1.h"
#include "layer2.h"
#include "layer2_proto.h"
#include "network.h"

sendpacket_t out_heads[OUT_BUFFER_SIZE];
l2_data_t *dm;
path_id sendpipe;
path_id recvpipe;

/* last time l2_ding() was called */
static u_int32 lasttime=0;


#define __DEBUG
#define DEFAULT_TIMEOUT 10 /* in secs */
/* make sure that out_heads only contain invalid process ID's 
 * return: E_OK
 */
int l2_init() {
	int lp1;
	u_int32 date;
	u_int16 day;
	u_int32 ticks;
		
	_os9_getime(1,&lasttime,&date,&day,&ticks);
	
	for (lp1=0;lp1<OUT_BUFFER_SIZE;lp1++)
		out_heads[lp1].process=0;
	return E_OK;
}


/* Called when a new frame is present in the in-buffer. Will
 * decrement in_changed if it manages to write the frame data to the recvpipe.
 * If it doesn't write the frame received, it will be called on next frame recv.
 * and try again.
 * return: E_OK
 *         E_PIPEOVERFLOW
 */
int l2_handle_inframe() {
	u_int32 pipecount, len;
	frame_t *f;
	recvpacket_t ph;
	
	/* only if someone is listening,  we will try to write the new frame to the pipe */
	if (dm->recv_listen) {
		/* peek the pipe to get the number of free bytes */
		_os_gs_ready(recvpipe, &pipecount);
		pipecount = RECV_PIPE_SIZE-pipecount;
		/* get pointer to last frame in the inbuffer */
		l1_peekInFrame(&f);
		
		if (f->size+sizeof(recvpacket_t) <= pipecount ) {
			/* unpack frame into packet */
			ph.signature=PACKETHEADSIGNATURE;
			ph.source=f->source;
			ph.dest=f->dest;
			ph.size=f->size;
			/* write packet (header+data) to recvpipe */
			len=sizeof(ph);
			_os_write(recvpipe, &ph, &len);
			len=ph.size;
			_os_write(recvpipe, &(f->data), &len);
			l1_doneInFrame();
			return E_OK;
		} else 
			return E_PIPEOVERFLOW;
	} else {
		/* if no-one is listening, we will discard the frame received and
		 * pretend everything went OK in layer2 */
		l1_doneInFrame();
		return E_OK;
	}
}

/* Whenever out_states are updated, this function is called and will send 
 * signals to all affected processes.
 * return: E_OK
 */
int l2_handle_replies() {
	int lp1;
	
	for (lp1=0;lp1<OUT_BUFFER_SIZE;lp1++) {
		switch (l1_getFrameStatus(lp1)) {
			case OUT_STATE_NAK:
				/* retransmit lost packet a number of times - NOT IMPLEMENTED */
				/* signal NAK */
#ifdef DEBUG
				printf("l2_handle_replies(): NAK'ed bufferno %d\n",lp1);
#endif
				if (out_heads[lp1].process)
					_os_send( out_heads[lp1].process, out_heads[lp1].signak);
				l1_freeFrame(lp1);
				break;

			case OUT_STATE_ACK:
				/* SIGNAL ACK */
#ifdef DEBUG
				printf("l2_handle_replies(): ACK'ed bufferno %d\n",lp1);
#endif
				if (out_heads[lp1].process)
					_os_send( out_heads[lp1].process, out_heads[lp1].sigack);
				l1_freeFrame(lp1);
				break;
			default:
				/* do nothing if not ack/nak */
				break;
		}
	}
	out_changed=0;
	return E_OK;
}

/* When data is available in the sendpipe, this function is called and will
 * read data from the sendpipe and add to outbuffer.
 * return: E_OK - when packet is added to outbuffer
 *         E_NOFREEBUFFERS - when no free buffers where found (already NAKed)
 *         E_NOTREADY - pipe data too small
 */
int l2_handle_new() {
	u_int32 pipecount;
	frame_t f;
	int indx;
	sendpacket_t ph;
		
	/* peek the sendpipe to get the number of bytes in use (avoid blocking) */
	_os_gs_ready(sendpipe, &pipecount);
	if ( pipecount<sizeof(sendpacket_t) ) 
		return E_NOTREADY;
#ifdef DEBUG
	printf("l2_handle_new(): found %i bytes total in pipe\n",pipecount);
#endif
	pipecount=sizeof(sendpacket_t);
	_os_read(sendpipe,(void*)&ph,&pipecount);
#ifdef DEBUG
	printf("l2_handle_new(): header: proc=%d ack=%d nak=%d to=%d, bytes=%d\n", ph.process, ph.sigack, ph.signak, ph.dest, ph.size);
#endif
	pipecount=ph.size;
	_os_read(sendpipe,(void*)&f.data,&pipecount);
	f.dest=ph.dest;
	f.size=pipecount;
#ifdef DEBUG
	printf("l2_handle_new(): read %d bytes...\n",pipecount);
#endif
	
	indx = l1_sendFrame(&f); 
	if (indx<0) { /* if unable to send */
		_os_send( ph.process, ph.signak );
		return indx;
	}
	out_heads[indx] = ph;
	out_heads[indx].timeout = DEFAULT_TIMEOUT;
#ifdef DEBUG
	printf("l2_handle_new(): got entry %d in outbuffer...\n",indx);
#endif
	return E_OK;
}


/* Retires packet with time-outs, messes with out_buffer, out_states
 * and out_changed directly... It is assumed, that this method is called
 * at least 2 times a day (otherwise it will miscalculate time passed and
 * severly mess up).
 * param :
 * return: E_OK
 */
int l2_ding() {
	u_int32 date;
	u_int16 day;
	u_int32 ticks;
	u_int32 time;
	int delta; /* time since last ding... */
	int lp1=0;
	
	_os9_getime(1,&time,&date,&day,&ticks);
	delta=(int)time-(int)lasttime;
	/* correct delta for julian time wrap-around */
	if (delta<0) 
		delta+=86400;
	/* update buffers if more than one sec. passed */
	if (delta) {
		lasttime=time;
		for (lp1=0;lp1<OUT_BUFFER_SIZE; lp1++) {
			if (out_state[lp1]==OUT_STATE_TRANS) {
				out_heads[lp1].timeout-=delta;
				if ( out_heads[lp1].timeout<=0 ) {
#ifdef DEBUG
					printf("l2_ding(): frame with bufferno %d timed out!\n",lp1);
#endif
					out_state[lp1]=OUT_STATE_NAK;
					out_changed++;
				}
			}
		}
	}
	return E_OK;
}
