/* File: testnet.c
 * This program tests the network driver and its interface.
 *
 */

#include <string.h>
#include <process.h>
#include <cglob.h>
#include <types.h>
#include <semaphore.h>
#include <module.h>
#include <modes.h>
#include <sg_codes.h>
#include <signal.h>
#include "net.h"
#include "layer2.h"

#define SIG_ACK 10000
#define SIG_NAK 10001
#define SIG_PIPEDATA 10002

#define QUEUE_SIZE 8

#define __DEBUG

/* signal counters */
int sig_term;
int sig_ack;
int sig_nak;
int sig_unknown;
int sig_pipedata;

/* internal variables for resource management */
static char ok_sighandler;
static char ok_dm;
static char ok_pipe_send;
static char ok_pipe_recv;
static char ok_sema_recv;
static char ok_sig_recv;
/* count number of unreplied writes done asynch. Must be 0 before terminating */
static int nowrites;
/* internal vars. for interacting with network */
static l2_data_t *dm;
static mh_com *dm_head;
static path_id sendpipe, recvpipe;

/* required for interfacing to network... */
void onSignal(signal_code sig) {
#ifdef DEBUG
	printf("net/onSignal: sigcode=%d\n",sig);
#endif
	switch (sig) {
		case 2: case 3:
			sig_term++;
			break;
	
		case SIG_ACK:
			sig_ack++;
			break;
		
		case SIG_NAK:
			sig_nak++;
			break;
		
		case SIG_PIPEDATA:
			sig_pipedata++;
			break;
		
		default:
			sig_unknown++;
			break;
	}
	
	_os_rte();
}

/* connect and initialize networking on level 2.
 * return: ENET_OK - no errors
 *         ENET_RECVPIPEFAIL - could not connect to recv. pipe
 *         ENET_SENDPIPEFAIL - could not connect to send pipe
 *         ENET_DATAMODULEFAIL - failed connecting to data module
 *         ENET_SIGHANDLERFAIL - couldn't install sighandler
 */
int net_init() {
	int errno = ENET_OK;
	u_int16 typelang;
	u_int16 attrrev;
	char dm_name[64];
	char *dm_namep = dm_name;

	/* reset resource management */
	ok_sighandler = 0;
	ok_dm = 0;
	ok_pipe_send = 0;
	ok_pipe_recv = 0;
	ok_sema_recv = 0;
	ok_sig_recv = 0;
	/* reset signal counters */
	sig_term=0;
	sig_ack=0;
	sig_nak=0;
	sig_unknown=0;
	sig_pipedata=0;
	nowrites=0;
	
	/* connect to the shared data module for and pipe-name */
#ifdef DEBUG
	printf("net/net_init(): linking to data module...\n");
#endif
	strcpy(dm_name,DATA_MODULE_NAME);
	typelang = mktypelang(MT_ANY, ML_ANY);
	attrrev = mkattrevs(MA_REENT, 0);
	if ( _os_link(&dm_namep, &dm_head, (void**) &dm, &typelang, &attrrev) )
		errno=ENET_DATAMODULEFAIL;
	else
		ok_dm=1;

	/* set up signal handler */
	if (ok_dm) {
#ifdef DEBUG
		printf("net/net_init(): setting up signal handler...\n");
#endif
		if ( _os_intercept(onSignal,_glob_data) )
			errno=ENET_SIGHANDLERFAIL;
		else
			ok_sighandler=1;
	}		

	/* get exclusive access to recv. pipe */
	if (ok_dm) {
#ifdef DEBUG
		printf("net/net_init(): locking semaphore...\n");
#endif
		if ( _os_sema_p( &(dm->recvpipe_sem) ) ) 
			errno=ENET_RECVPIPEFAIL;
		else {
			ok_sema_recv=1;
			dm->recv_listen=1;
		}
	}
	
	/* open recv. pipe */
	if (ok_sema_recv) {
#ifdef DEBUG
		printf("net/net_init(): attaching to receive pipe\n");
#endif
		if ( _os_open(dm->recvpipe, FAM_READ, &recvpipe) )
			errno=ENET_RECVPIPEFAIL;
		else 
			ok_pipe_recv=1;
			
		/* set up a signal to catch when receiving data */
		if ( _os_ss_sendsig(recvpipe, SIG_PIPEDATA) )
			errno=ENET_RECVPIPEFAIL;
		else
			ok_sig_recv=1;
	}
	
	/* open the send-pipe */
	if (ok_dm && ok_pipe_recv) {
#ifdef DEBUG
		printf("net/net_init(): attaching to send pipe...\n");
#endif
		if ( _os_open(dm->sendpipe, FAM_WRITE, &sendpipe) )
			errno=ENET_SENDPIPEFAIL;
		else 
			ok_pipe_send=1;
	}
	
	if (errno) {
#ifdef DEBUG
		printf("net/net_init(): closing network due to error %d...\n",errno);
#endif
		net_close();
	} else {
#ifdef DEBUG
		printf("net/net_init(): done %d...\n",errno);
#endif
	}
	return errno;
}


/* Termiates the use of the network
 * return: ENET_OK - no errors
 *         ENET_RECVPIPEFAIL - could not connect to recv. pipe
 *         ENET_SENDPIPEFAIL - could not connect to send pipe
 *         ENET_DATAMODULEFAIL - failed connecting to data module
 *         ENET_SIGHANDLERFAIL - couldn't install sighandler
 */
int net_close() {
	int errno= ENET_OK;
	
	dm->recv_listen=0;
	
	/* wait till replies are recv. */
/*	while (nowrites)
		if (net_wait(0)==ENET_TERM)
			break;
*/	
	/* free recv-pipe */
	if (ok_pipe_recv) {
		if (_os_close(recvpipe))
			errno = ENET_RECVPIPEFAIL;
		else 
			ok_pipe_recv=0;
	}
	/* release the receive pipe semaphore */
	if (ok_sema_recv) {
		if ( _os_sema_v(&(dm->recvpipe_sem)) )
			errno = ENET_RECVPIPEFAIL;
		else 
			ok_sema_recv=0;
	}
	
	/* free send-pipe */
	if (ok_pipe_send) {
		if (_os_close(sendpipe)) 
			errno=ENET_SENDPIPEFAIL;
		else 
			ok_pipe_send=0;
	}
	
	/* free shared data module */
	if (ok_dm) {
		if ( _os_unlink(dm_head) ) 
			errno=ENET_DATAMODULEFAIL;
		else 
			ok_dm=0;
	}
	/* don't care to release signal handler */
	return errno;
}
	

/* Sends data over network to destination. Call net_wait to wait for response
 * (or new packets).
 * return: ENET_OK - successfull delivery
 *         ENET_NOTREADY - network not ready
 *         ENET_INVALIDPACKETSIZE - packet size outside 0..254
 *         ENET_INVALIDDEST - destination outside 0..15
 *         ENET_PIDFAILED - could not get process ID
 *         ENET_SENDPIPEFAIL - could not obtain exclusive access to sendpipe
 */
int net_send(char dest, void* data, int len) {
	sendpacket_t packet_head;
	u_int16 priority, user, group;
	u_int32 count;
	char buffer[300];
	
	if (!ok_pipe_send)
		return ENET_NOTREADY;
	if (len<0 || len>254) 
		return ENET_INVALIDPACKETSIZE;
	if (dest<0 || dest>15)
		return ENET_INVALIDDEST;
	if (_os9_id(&packet_head.process,&priority,&group,&user))
		return ENET_PIDFAILED;

	if (ok_pipe_send) {
		if ( _os_sema_p( &(dm->sendpipe_sem) ) ) 
			return ENET_SENDPIPEFAIL;
		else {
			packet_head.signature = PACKETHEADSIGNATURE;
			/* obtain process ID etc. */
			packet_head.sigack=SIG_ACK;
			packet_head.signak=SIG_NAK;
			packet_head.timeout=100;
			packet_head.dest = dest;
			packet_head.size=len;
			memcpy(buffer,&packet_head,sizeof(packet_head));
			memcpy(buffer+sizeof(packet_head),data,len);
			count=sizeof(packet_head)+len;
			_os_write( sendpipe,(void*)buffer,&count);
			_os_sema_v( &(dm->sendpipe_sem) );
			nowrites++;
		}
		
	}
}


/* Waits for data to receive, reads from comm.pipe.
 * param : buffer - pointer to memory location to store frame
 *         len - pointer to buffer len, will be updated to real
 *               length of packet. 
 * return: ENET_OK - successfull delivery
 *         ENET_NOTREADY - network not setup!
 *         ENET_INVALIDPACKETSIZE - packet size outside 0..254
 *         ENET_INVALIDDEST - destination outside 0..15
 *         ENET_PIDFAILED - could not get process ID
 *         ENET_RECVPIPEFAIL - could not obtain exclusive access to sendpipe
 *         ENET_BUFFERTOOSMALL - buffer is too small to hold data
 *         ENET_NULLPOINTER - buffer or len is null-pointer
 *         ENET_NOPACKETS - no packets waiting in pipe
 */
int net_recv(void* buffer, int *len, int* from, int* to) {
	static packetinpipe = 0;
	static recvpacket_t phead; 
	u_int32 headlen;
	
	if (!ok_pipe_recv || !ok_sema_recv || !ok_sighandler)
		return ENET_NOTREADY;
	if (!buffer || !len || !from || !to) 
		return ENET_NULLPOINTER;
	if (!sig_pipedata) 
		return ENET_NOPACKETS;
	
	if (!packetinpipe) {
		headlen=sizeof(recvpacket_t);
		_os_read( recvpipe, &phead, &headlen );
	}
#ifdef DEBUG
	printf("net/net_read():read packet from pipe %i bytes\n",headlen);
#endif
		
	if (phead.size>*len) {
		packetinpipe=1;
		return ENET_BUFFERTOOSMALL;
	} else 
		packetinpipe=0;
	headlen=phead.size;
	_os_read( recvpipe,buffer,&headlen );
#ifdef DEBUG
	printf("net/net_read(): read packet of %i bytes\n",headlen);
#endif

	*len=headlen;

#ifdef DEBUG
	_os_gs_ready(recvpipe, &headlen);
	printf("net/net_read(): found %i bytes more ready in pipe\n",headlen);
#endif
	
	*from = phead.source;
	*to = phead.dest;
	
	/* receive signal next time data is available in pipe */
	sig_pipedata--;
	if (_os_ss_sendsig(recvpipe, SIG_PIPEDATA))
		return ENET_RECVPIPEFAIL;
	
	return ENET_OK;
}	

/* Waits for network events. Returns why it terminated waiting...
 * param : ticks - number of ticks to pass to _os9_sleep()
 *                 0 is indef, 1 is timeslice, 2.. nofticks
 * return: ENET_UNKNOWNSIGNAL - got a signal for you to handle
 *         ENET_INCOMING - new frame received 
 *         ENET_NAK - got NAK for last frame transmittet
 *         ENET_ACK - got ACK for last frame transmittet
 *         ENET_TERM - terminate signals arrived
 */
int net_wait(u_int32 ticks) {
	int errno;
	u_int32 i_ticks;
	
	/* wait for ACK/NAK on packet (or new packets) */
	i_ticks=ticks;
	_os_sigmask(1);
	if ( !sig_ack && !sig_nak && !sig_term && !sig_unknown && !sig_pipedata)
		_os9_sleep(&i_ticks);
	_os_sigmask(0);
		
	if (sig_term) 
		return ENET_TERM;
	
	if (sig_ack) {
		nowrites--;
		sig_ack--;
		return ENET_ACK;
	}

	if (sig_nak) {
		nowrites--;
		sig_nak--;
		return ENET_NAK;
	}

	if (sig_pipedata)  {
		return ENET_INCOMING;
	}
	
	if (sig_unknown)
		sig_unknown--;
	return ENET_UNKNOWNSIGNAL;
}


#include <stdio.h>
#include <stdlib.h>
/* Pass single character to this function to get the value of the
 * digit. This is done by looking up the value in a list.
 * If a negative value is returned, the character is not a hex-digit.
 */
int hexval(char c) {
	int lp1=0;
	char hexdig[16]="0123456789ABCDEF";
	
	lp1=0;
	while (lp1<16 && c != hexdig[lp1]) lp1++;
	if (lp1>15) return -1;
	else return lp1;
}


int main(int argc, char* argv[]) {
	int dest;
	int errno;
	int s,d;
	char buffer[512];
	char *bufp=buffer;
	int bsz;
	int outstanding[QUEUE_SIZE];
	int oss=0, ose=0;
	
	/* check for arguments */
	if (argc < 2) {
		printf("Usage: %s <our_special_sequence_format>\n\tThe format for the message is: <target>{<targets>}:<msg>\nwhere target is 1..F (1 char)", argv[0]);
		_os_exit(1);
	}
	
	/* clear buffer */
	for (bsz=0;bsz<sizeof(buffer);bsz++) 
		buffer[bsz]=0;
	/* copy message to buffer and parse addr */
	strcpy(buffer,argv[1]);
	bufp=buffer;
	dest=hexval(*bufp);
	bufp++;
	bsz=strlen(bufp)+1;

	if (dest<1 || dest>15) {
		printf("%s: invalid first target %d...\n", argv[0],dest);
		_os_exit(1);
	}
	printf("%s: the message to: %d is '%s'...\n", argv[0],dest,bufp);

	printf("%s: initing network...\n",argv[0]);
	switch ( errno=net_init() ) {
		case ENET_OK: break;
		case ENET_RECVPIPEFAIL: printf("%s: could not connect to recv. pipe\n", argv[0]); break;
		case ENET_SENDPIPEFAIL: printf("%s: could not connect to send pipe\n", argv[0]); break;
		case ENET_DATAMODULEFAIL: printf("%s: failed connecting to data module\n", argv[0]); break;
		case ENET_SIGHANDLERFAIL: printf("%s: couldn't install sighandler\n", argv[0]); break;
		default: printf("%s: unknown error %d\n", argv[0],errno);
	}
	if (errno) 
		_os_exit(1);
	
	printf("%s: transmitting message '%s'...\n",argv[0], bufp);
	switch ( errno=net_send(dest,(void*)bufp, bsz) ) {
		case ENET_OK: 
			outstanding[oss++]=dest;
			break;
		case ENET_RECVPIPEFAIL: printf("%s: could not connect to recv. pipe\n", argv[0]); break;
		case ENET_SENDPIPEFAIL: printf("%s: could not connect to send pipe\n", argv[0]); break;
		case ENET_DATAMODULEFAIL: printf("%s: failed connecting to data module\n", argv[0]); break;
		case ENET_SIGHANDLERFAIL: printf("%s: couldn't install sighandler\n", argv[0]); break;
		default: printf("%s: unknown error\n", argv[0]);
	}
	if (errno) _os_exit(1);
	printf("%s: waiting for response...\n",argv[0]);

	
	do {
		switch ( errno=net_wait(000) ) {
			case ENET_INCOMING:
				bsz=sizeof(buffer);
				errno=net_recv((void*)buffer, &bsz, &s, &d);
				if ( errno==ENET_OK ) {
					if (bsz>=0) {
						bufp=buffer;
						printf("%s: (%i->%i), %i bytes: '%s'...\n", argv[0], s, d, bsz, buffer);
						dest=hexval(*bufp);
						if (dest>0 && bsz>=0) {
							bufp++;
							bsz--;
							errno=net_send(dest, bufp, bsz);
							if (errno) 
								printf("%s: could not SEND the message to %d (e=%d)...\n", argv[0],dest,errno);
							else {
								outstanding[oss++]=dest; /* put dest in queue */
								if (oss>=QUEUE_SIZE) oss=0; /* wraparound */
								printf("%s: SENT the message to %d...\n", argv[0],dest);
							}
						} else 
							printf("%s: '%s' path ended...\n", argv[0], buffer);
					} else 
						printf("%s: path ended with 0 bytes...\n", argv[0]);
				} else
					printf("%s: error receiving....\n", argv[0]);
				
				break;
				
			case ENET_NAK:
				dest = outstanding[ose++]; /* remove dest from queue */
				if (ose>=QUEUE_SIZE) ose=0; /* wraparound */
				printf("%s: packet to %d not delivered!\n", argv[0], dest); 

				dest=hexval(*bufp);
				if (dest>0 && dest<16 && bsz>=0) {
					bufp++;
					bsz--;
					printf("%s: trying %d instead...\n", argv[0], dest); 
					errno=net_send(dest, bufp, bsz);
					if (errno) 
						printf("%s: could not send the message to %d (e=%d)...\n", argv[0],dest,errno);
					else {
						outstanding[oss++]=dest; /* put dest in queue */
						if (oss>=QUEUE_SIZE) oss=0; /* wraparound */
						printf("%s: send the message to %d...\n", argv[0],dest);

					}

				} else {
					printf("%s: '%s' path ended...\n", argv[0], bufp);
				}
/*				printf("%s: packet not delivered to %d...\n", argv[0], dest); 
*/				break;
				
			case ENET_ACK:
				dest = outstanding[ose++]; /* remove dest from queue */
				if (ose>=QUEUE_SIZE) ose=0; /* wraparound */
				printf("%s: packet delivered to %d correctly...\n", argv[0], dest);
				break;
		}
	} while (!sig_term && !sig_unknown);
	
	printf("%s: closing network...\n",argv[0]);
	switch ( errno=net_close() ) {
		case ENET_OK: break;
		case ENET_RECVPIPEFAIL: printf("%s: could not disconnect from recv. pipe\n", argv[0]); break;
		case ENET_SENDPIPEFAIL: printf("%s: could not disconnect from send pipe\n", argv[0]); break;
		case ENET_DATAMODULEFAIL: printf("%s: failed unlinking from data module\n", argv[0]); break;
		case ENET_SIGHANDLERFAIL: printf("%s: couldn't uninstall sighandler\n", argv[0]); break;
		default: printf("%s: unknown error\n", argv[0]);
	}
	if (errno) _os_exit(1);
	_os_exit(0);
}



