/* File: net.h
 * 
 * Defines error codes and prototypes for functions relating to networking
 * When writing an application using the network, use net_wait() and react
 * corrospondingly to the return code received from net_wait - that means
 * call functions to receive packets immediately (and use net_wait() as
 * often as possible).
 *
 * The usual sequence for networking is:
 *   net_send(...)
 *   switch (net_wait(0)) {
 *     case ...:
 *
 * Internally, a signal is sent, which is handled in the onSignal(). Therefore,
 * never mess with signal-handlers (or at least make sure our signal-handler
 * is called on signals. The signal range 1000-1016 is reserved for network
 * use.
 *
 */

#ifndef NET_H
#define NET_H

#include <types.h>

#define ENET_OK 0 /* no errors */
#define ENET_RECVPIPEFAIL -1 /* could not connect to recv. pipe */
#define ENET_SENDPIPEFAIL -2 /*could not connect to send pipe */
#define ENET_DATAMODULEFAIL -3 /* failed connecting to data module */
#define ENET_SIGHANDLERFAIL -4 /* couldn't install sighandler */
#define ENET_INVALIDPACKETSIZE -5 /* packet size outside 0..254 */
#define ENET_INVALIDDEST -6 /* destination outside 0..15 */
#define ENET_PIDFAILED -7 /* could not get process ID */
#define ENET_BUFFERTOOSMALL -8 /* buffer is too small to hold data */
#define ENET_NULLPOINTER -9 /* buffer or len is null-pointer */
#define ENET_NOPACKETS -10 /* no packets waiting in pipe */
#define ENET_INCOMING -11 /* new frame received */
#define ENET_NAK -12 /* got NAK for last frame transmittet */
#define ENET_ACK -13 /* got ACK for last frame transmittet */
#define ENET_TERM -14 /* terminate signals arrived */
#define ENET_UNKNOWNSIGNAL -15 /* got a signal for you to handle */
#define ENET_NOTREADY -16 /* network not ready */

/* when awakened from a sleep, check these signal counters to see why you
 * were awakened. Usually, you would use net_wait() to sleep (indefinitely)
 */
extern int sig_term; /* ctrl-c or ctrl-e */
extern int sig_unknown; /* received unknown signal, usually quit (remember net_close() ) */
extern int sig_ack; /* ctrl-c or ctrl-e */
extern int sig_nak; /* received unknown signal, usually quit (remember net_close() ) */

/* connect and initialize networking on level
 * return: ENET_OK - no errors
 *         ENET_RECVPIPEFAIL - could not connect to recv. pipe
 *         ENET_SENDPIPEFAIL - could not connect to send pipe
 *         ENET_DATAMODULEFAIL - failed connecting to data module
 *         ENET_SIGHANDLERFAIL - couldn't install sighandler
 */
int net_init(void);

/* Termiates the use of the network
 * return: ENET_OK - no errors
 *         ENET_RECVPIPEFAIL - could not connect to recv. pipe
 *         ENET_SENDPIPEFAIL - could not connect to send pipe
 *         ENET_DATAMODULEFAIL - failed connecting to data module
 *         ENET_SIGHANDLERFAIL - couldn't install sighandler
 */
int net_close(void);

/* Sends data over network to destination. Call net_wait to wait for response
 * (or new packets).
 * return: ENET_OK - successfull delivery
 *         ENET_NOTREADY - network not ready
 *         ENET_INVALIDPACKETSIZE - packet size outside 0..254
 *         ENET_INVALIDDEST - destination outside 0..15
 *         ENET_PIDFAILED - could not get process ID
 *         ENET_SENDPIPEFAIL - could not obtain exclusive access to sendpipe
 */
int net_send(char dest, void *data, int len);

/* Waits for data to receive, reads from comm.pipe.
 * param : buffer - pointer to memory location to store frame
 *         len - pointer to buffer len, will be updated to real
 *               length of packet. 
 * return: ENET_OK - successfull delivery
 *         ENET_NOTREADY - network not setup!
 *         ENET_INVALIDPACKETSIZE - packet size outside 0..254
 *         ENET_INVALIDDEST - destination outside 0..15
 *         ENET_PIDFAILED - could not get process ID
 *         ENET_RECVPIPEFAIL - could not obtain exclusive access to sendpipe
 *         ENET_BUFFERTOOSMALL - buffer is too small to hold data
 *         ENET_NULLPOINTER - buffer or len is null-pointer
 *         ENET_NOPACKETS - no packets waiting in pipe
 */
int net_recv(void* buffer, int *len, int* from, int* to);

/* Waits for network events. Returns why it terminated waiting...
 * param : ticks - number of ticks to pass to _os9_sleep()
 *                 0 is indef, 1 is timeslice, 2.. nofticks
 * return: ENET_UNKNOWNSIGNAL - got a signal for you to handle
 *         ENET_INCOMING - new frame received 
 *         ENET_NAK - got NAK for last frame transmittet
 *         ENET_ACK - got ACK for last frame transmittet
 *         ENET_TERM - terminate signals arrived
 */
int net_wait(u_int32 ticks);

#endif
