/* File: network.c
 * This file performs the actual interaction with the driver and contains
 * the main() for all parts.
 */
#include <stdio.h>
#include <string.h>
#include <types.h>
#include <process.h>
#include <sg_codes.h>
#include <modes.h>
#include <signal.h>
#include <module.h>
#include <semaphore.h>
#include <time.h>
#include <cglob.h>  /* Definition of *_glob_data required by '_os_intercept' */
#include <stdlib.h> /* for atoi() */

#include "status.h"
#include "layer1.h"
#include "layer2.h"
#include "layer2_proto.h"
#include "network.h"
#include "drvinterface.h"


#define SIG_DATATOSEND 1000
static int sig_read;
static int sig_write;
static int sig_sendpipe;
static int sig_term;
static int sig_unknown;

/* the buffers for the driver interface */
#define DRVR_INBUFLEN 1024
char drvr_in_buffer[DRVR_INBUFLEN];
char drvr_out_buffer[1024];

void onSignal(signal_code s) {
	switch (s) {
		case 2: case 3: 
			sig_term++;
			break;
			
		case SIG_DONEWRITE:
			sig_write++;
			break;
		
		case SIG_DONEREAD:
			sig_read++;
			break;
		
		case SIG_DATATOSEND:
			sig_sendpipe++;
			break;

		default:
			sig_unknown++;
	}
	_os_rte();
}	

int run(void) {
	u_int32 ticks;
	int len = 0;
	int pendingWrite = 0;
	char *drvr_buf_pos = drvr_in_buffer;
	int max_sigread=0;
	int nrecv=0;
	int nforward=0;
	int nsent=0;
	int e=0;
	u_int32 timeout;
	u_int32 date, time;
	u_int16 day;
	
	_os_sigmask(1);
	sig_sendpipe=sig_unknown=sig_term=sig_read=sig_write=0;
	_os_sigmask(0);
	
	StartBufferRead( drvr_in_buffer, sizeof(drvr_in_buffer) );
	
	do {
		/* sleep until signaled */
		ticks=2*CLOCKS_PER_SEC;
		_os_sigmask(1);
		if (!sig_read && !sig_write && !sig_sendpipe && !sig_term)
			_os9_sleep(&ticks);
		else
			_os_sigmask(0);
	
		/* determine max. sigreads (number of bytes we are behind in the 
		 * drivers buffer */
		if (sig_read>max_sigread) max_sigread=sig_read;
		
		/* character received */
		while (sig_read) {
			l1_onReadChar( *(drvr_buf_pos++) );
			if ( drvr_buf_pos>=(drvr_in_buffer+DRVR_INBUFLEN) )
				drvr_buf_pos=drvr_in_buffer;
			sig_read--;
			/* if we added something to the forward buffer, break the loop */
			if (forward_changed) 
				break;

		}
		/* transmission of frame completed (or no pending write) */
		if (sig_write || !pendingWrite) {
			if (sig_write) {
				sig_write--;
				pendingWrite=0;
			}
			/* if frames are ready for forwarding, then start next forward */
			if ( len=l1_prepareForward(drvr_out_buffer) ) {
				nforward++;
				StartBufferWrite( drvr_out_buffer, len );
				pendingWrite=1;
				forward_changed--;
			} else if (len=l1_prepareOutBuffer(drvr_out_buffer)) {
				nsent++;
				StartBufferWrite( drvr_out_buffer, len );
				pendingWrite=1;
			}
		}

		/* if something changed in the in-buffer, call layer2's for handling 
		 * incoming frames (are in in_buffer) */
		if (in_changed)
			if (l2_handle_inframe()==E_OK) {
				nrecv++;
				in_changed--;
			}

		/* give layer2 a chance to retire frames... */
		l2_ding();
		
		/* if some states changed in the out_buffer, let layer2 handle the 
		 * changes. l2_handle_replies will handle ALL changes. */
		if (out_changed) {
			l2_handle_replies();
			out_changed=0;
		}
		
		/* if someone added data to the sendpipe, let layer2 handle it */
		if (sig_sendpipe) {
			e=l2_handle_new();
			sig_sendpipe--;
			if (e==E_OK) {	
				if (!pendingWrite) {
					if ( len=l1_prepareForward(drvr_out_buffer) ) {
						nforward++;
						StartBufferWrite( drvr_out_buffer, len );
						pendingWrite=1;
						forward_changed--;
					} else if (len=l1_prepareOutBuffer(drvr_out_buffer)) {
						nsent++;
						StartBufferWrite( drvr_out_buffer, len );
						pendingWrite=1;
					}
				}
				/* get a signal next time data is in the pipe */
				_os_ss_sendsig(sendpipe, SIG_DATATOSEND);
			}
		}
	} while (!sig_unknown && !sig_term);

	printf("frames sent: %i\nframes received: %i\nframes forwarded: %i\nmax_sigread: %i\n", nsent, nrecv, nforward, max_sigread);	
}



int main(int argc, char* argv[]) {
	char ok_driver = 0;
	char ok_sighandler = 0;
	char ok_dm = 0;
	char ok_sema_send = 0;
	char ok_sema_recv = 0;
	char ok_pipe_send = 0;
	char ok_pipe_recv = 0;
	char ok_sig_send = 0;
	mh_data *dm_head;
	u_int16 typelang;
	u_int16 attrrev;
	int netaddr;
	process_id pid;
	u_int16 priority, age, group, user, newpri;
	int32 schedule;
	
	if ((argc < 2)) {
		printf("Usage: %s <network-address (range 1-15)> [<priority>=norm+128]\n", argv[0]);
		_os_exit(1);
	}
	netaddr=atoi(argv[1]);
	if (netaddr>15 || netaddr<1){
	 	printf("Invalid netaddres!\n");
		_os_exit(1);
	}
		
	/* increase our priority */
	_os_id(&pid, &priority, &age, &schedule, &group, &user);
	if (argc>2)
		newpri=atoi(argv[2]);
	else
		newpri=priority+128;
	_os_setpr(pid, newpri);
	
	/* connect to the driver */
	if ( InitDrv() ) {
		printf("%s: connect to driver failed, exiting!\n", argv[0]);
		_os_exit(1);
	} else {
		printf("%s: connected to driver...\n", argv[0]);
		ok_driver=1;
	}

	/* create shared data module for semaphore */
	if (ok_driver) {
		typelang = mktypelang(MT_ANY, ML_ANY);
		attrrev = mkattrevs(MA_REENT, 0);
		if ( _os_datmod(DATA_MODULE_NAME, sizeof(l2_data_t), &attrrev, &typelang, MP_OWNER_READ|MP_OWNER_WRITE, (void**) &dm, &dm_head ) )
			printf("%s: failed creating shared data module!\n", argv[0]);
		else {
			printf("%s: created shared data module...\n", argv[0]);
			ok_dm=1;
		}
	}

	/* init semaphores */
	if (ok_dm) {
		dm->recv_listen =0;
		if ( _os_sema_init( &(dm->sendpipe_sem) ) ) 
			printf("%s: failed creating send-semaphore!\n", argv[0]);
		else {
			printf("%s: created send-semaphore...\n", argv[0]);
			ok_sema_send=1;
		}
		if ( _os_sema_init( &(dm->recvpipe_sem) ) ) 
			printf("%s: failed creating recv-semaphore!\n", argv[0]);
		else {
			printf("%s: created recv-semaphore...\n", argv[0]);
			ok_sema_recv=1;
		}
	}
	
	/* create send-pipe, setup signals */
	if (ok_sema_send) {
		strcpy(dm->sendpipe, SEND_PIPE_NAME);
		if ( _os9_create(dm->sendpipe, FAM_READ|FAM_SIZE, &sendpipe, FAP_READ|FAP_WRITE|FAP_PWRITE, SEND_PIPE_SIZE) )
			printf("%s: failed creating send pipe!\n", argv[0]);
		else {
			printf("%s: created send pipe...\n", argv[0]);
			ok_pipe_send=1;
		}
		
		if ( _os_ss_sendsig(sendpipe, SIG_DATATOSEND) )
			printf("%s: failed setting up signal-on-data in send-pipe!\n", argv[0]);
		else
			ok_sig_send=1;
	}
	
	/* create recv-pipe, setup signals */
	if (ok_sema_recv) {
		strcpy(dm->recvpipe, RECV_PIPE_NAME);
		if ( _os9_create(dm->recvpipe, FAM_WRITE|FAM_SIZE, &recvpipe, FAP_PREAD|FAP_WRITE|FAP_READ,RECV_PIPE_SIZE) ) 
			printf("%s: failed creating receive pipe!\n", argv[0]);
		else {
			printf("%s: created receive pipe...\n", argv[0]);
			ok_pipe_recv=1;
		}
	}





	if (ok_sig_send && ok_sema_recv && ok_sema_send && ok_pipe_recv && ok_pipe_send && ok_dm ) {
		_os_intercept(onSignal,_glob_data);
		/* initialize network */
		l1_init(netaddr);
		l2_init();
		/* do actual processing here... */
		printf("%s: awaiting data on address %i...\n", argv[0], netaddr);
		run();
	} else 
		printf("%s: Conditions for running _NOT_ OK!\n", argv[0]);
	
	
	
	
	
	/* free recv-pipe */
	if (ok_pipe_recv) {
		if (_os_close(recvpipe)) 
			printf("%s: failed closing recv pipe!\n", argv[0]);
		else {
			printf("%s: closed recv pipe...\n", argv[0]);
			ok_pipe_recv=0;
		}
	}
	
	
	/* free send-pipe */
	if (ok_pipe_send) {
		if (_os_close(sendpipe)) 
			printf("%s: failed closing send pipe!\n", argv[0]);
		else {
			printf("%s: closed send pipe...\n", argv[0]);
			ok_pipe_send=0;
		}
	}
	
	/* free semaphores */
	if (ok_sema_send) {
		if ( _os_sema_term( &(dm->sendpipe_sem)) ) 
			printf("%s: failed terminating send-semaphore!\n", argv[0]);
		else {
			printf("%s: send-semaphore terminated...\n", argv[0]);
			ok_sema_send=0;
		}
	}
	if (ok_sema_recv) {
		if ( _os_sema_term( &(dm->recvpipe_sem)) ) 
			printf("%s: failed terminating recv-semaphore!\n", argv[0]);
		else {
			printf("%s: recv-semaphore terminated...\n", argv[0]);
			ok_sema_recv=0;
		}
	}
		
	/* free shared data module */
	if ( ok_dm ) {
		if ( _os_unlink( (mh_com*) dm_head) ) 
			printf("%s: failed unlinking shared data module!\n", argv[0]);
		else {
			printf("%s: unlinked shared data module...\n", argv[0]);
			ok_dm=0;
		}
	}
	/* disconnect from driver */
	if (ok_driver) {
		if ( CloseDrv() ) 
			printf("%s: failed disconnecting from driver!\n", argv[0]);
		else {
			printf("%s: driver closed...\n", argv[0]);
			ok_dm=0;
		}
	}
		
	
	_os_exit(0);
}
