/* Motorola S-record reader, rev. 0.0.1
 * 2001-06-12 Rune M. Barnkob
 *
 * This is a single-character reader - it continues to look for 'S' until it
 * finds it and assumes it is the beginning of a S-record. If the record has
 * errors, it signals the error and returns to looking for 'S' so as a correct
 * recover may be done in an embedded situation.
 */
#include <stdio.h>


/* Update these typedefs to match the compiler */
typedef unsigned int uint;
typedef unsigned char byte;
typedef unsigned char uint8;
typedef signed char sint8;
typedef unsigned short int uint16;
typedef signed short int int16;
typedef signed int int32;
typedef unsigned int uint32;

/* Error codes */
#define E_SUCCESS 0 /* No error occured */
#define E_NOSRECORD -1 /* This is not an S-record */
#define E_CHECKSUM_ERROR -2 /* Checksum does not match */
#define E_LENGTH_MISMATCH -4 /* There is a mismatch in the length */
#define E_NOTHEX -5 /* character is not hex-digit */
#define E_ADDRESSERROR -6 /* Error occured while reading address */
#define E_DATAERROR -7 /* Data contained an errorneous character and was ignored */
#define E_EOF -8 /* End of data stream reached */
#define E_NOSUPPORT -9 /* This S-record type is not supported */

#define ACK 1
#define NAK 2
#define DEBUG1

/* Pass single character to this function to get the value of the
 * digit. This is done by looking up the value in a list.
 * If a negative value is returned, the character is not a hex-digit.
 */
int hexDigit(char c) {
	int lp1=0;
	char hexdig[16]="0123456789ABCDEF";
	
	lp1=0;
	while (lp1<16 && c != hexdig[lp1]) lp1++;
	if (lp1>15) return E_NOTHEX;
	else return lp1;
}


int charp = 0;
		
/* Reads single character off input. Returns negative if EOF (or none) */
int getChar() {
	int c = getchar();
	charp++;
	if (c==10 || c==13) charp=0;
#ifdef DEBUG2
	printf("getChar() = %c\n", c);
#endif
	return c;
}

/* Puts a character on the output (in real version). */
int putChar(int c) {
	switch (c) {
		case ACK: printf("<ACK> Acknowledged\n"); break;
		case NAK: printf("<NAK> Not Acknoledged - error in data\n"); break;
		default: printf("%c",c);
	}
}
	
/* checksummer: pass pointer to current checksum and update byte */
#define chk_upd(upd) global_checksum=(global_checksum + (upd) & 0xff) & 0xff
#define chk_zero global_checksum=0;
#define chk_get (~global_checksum & 0xff)
int global_checksum = 0;

/* Starts parsing an S-record. Repeats reading characters off the input
 * until it gets a 'S' (replies with NAK until ='S'). Then it expects
 * to read a character containing '0'..'9' (replies to putChar with NAK).
 * - mem must point to the global memory assigned to this reader.
 */
int parseRecs( byte* mem ) {
	int ch;
	int type;
	int len;
	int hi, lo;
	int e;
#ifdef DEBUG
	printf("parseRecs(%p)\n", mem);
#endif
	
	while ( (ch=getChar()) >0) {
		while (ch>0 && ch!='S') {
			ch = getChar();
/*			if (ch!='S' || ch!=13) putChar(NAK);
*/
		}
		if (ch<0) return E_EOF;
		type = getChar();
		hi = hexDigit(getChar()); if (hi<0) continue;
		lo = hexDigit(getChar()); if (lo<0) continue;
		len = hi<<4 | lo;	
		chk_zero; chk_upd(len);
		switch (type) {
			case 0: continue;
			case '0': e=parseS0(mem, len); break;
			case '1': e=parseMem(mem, len, 2); break;
			case '2': e=parseMem(mem, len, 3); break;
			case '3': e=parseMem(mem, len, 4); break;
			case '5': e=parseS5(mem, len); break;
			case '7': e=parseS7(mem, len); break;
			case '8': e=parseS8(mem, len); break;
			case '9': e=parseS9(mem, len); break;
		}
		if (e<0) putChar(NAK);
		else putChar(ACK);
		switch (e) {
			case E_SUCCESS: break;
			case E_NOSRECORD: printf("(%d) This is not an S-record\n",charp); break;
			case E_CHECKSUM_ERROR: printf("(%d) Checksum does not match\n",charp); break;
			case E_LENGTH_MISMATCH: printf("(%d) There is a mismatch in the length\n",charp); break;
			case E_NOTHEX: printf("(%d) character is not hex-digit\n",charp); break;
			case E_ADDRESSERROR: printf("(%d) Error occured while reading address\n",charp); break;
			case E_DATAERROR: printf("(%d) Data contained an errorneous character and was ignored\n",charp); break;
			case E_EOF: printf("(%d) End of data stream reached\n",charp); break;
			case E_NOSUPPORT: printf("(%d) This type of S-record is not supported\n",charp); break;
		}
	}
}

/* Parses a stream of bytes into buffer and confirms the checksum. If the
 * checksum mismatches or the reading fails, a negative value is returned.
 * It is assumed that buffer is large enough to hold all data
 *   buffer: buffer to read data into
 *   len: length of data to read
 */
int parseData(byte* buffer, int len) {
	int hi, lo, chk, rchk;
	int chi, clo;
	int ch;
	byte b;
#ifdef DEBUG
	printf("parseData(%p,%d)\n", buffer, len);
#endif

	/* Read data into buffer */
	while (len>1) {
		chi = getChar(); if (chi<0) return E_EOF;
		clo = getChar(); if (clo<0) return E_EOF;
		hi = hexDigit(chi); if (hi<0) return E_DATAERROR;
		lo = hexDigit(clo); if (lo<0) return E_DATAERROR;
		b = hi<<4 | lo; 
		chk_upd(b);
		*buffer = b;
		buffer++;
		len--;
	}
	
	/* read checksum from stream */
	hi = hexDigit(getChar()); if (hi<0) return E_DATAERROR;
	lo = hexDigit(getChar()); if (lo<0) return E_DATAERROR;
	rchk = hi<<4 | lo;
	
	/* verify checksum */
#ifdef DEBUG
	printf("\tChecksums: calc=%x, read=%x\n", chk_get, rchk);
#endif
	if (rchk!=chk_get) return E_CHECKSUM_ERROR;
	else return E_SUCCESS;
}
	

/* Parses an S0 type record. The values are printed on screen */
int parseS0( byte* mem, int len ) {
	byte buffer[256];
	int e;
	int lp1;
	int ch;
	uint32 address;
#ifdef DEBUG
	printf("parseS0(%p,%d)\n", mem, len);
#endif
	
	/* Read address to waste  */
	for (lp1=0; lp1<4; lp1++) {
		ch = hexDigit(getChar());
		if (ch<0) return E_ADDRESSERROR;
		address = (address<<4) | ch;
		if (lp1&1) chk_upd(address);
	}

	/* Parse the data into buffer and get the checksum of the data read */
	e = parseData(buffer, len-2);
	if (e<0) return e;
	buffer[len-3] = 0; /* null termination */
	
	/* if data matches print it on screen (or ignore it???) */
	printf("SO: %s\n", buffer);
	
	return E_SUCCESS;
}

/* Parses memory type S-record
 *   mem: memory to place result in
 *   len: length of data block
 *   addrsz: number of bytes address consumes
 */
int parseMem( byte* mem, int len, int addrsz ) {
	uint32 address=0;
	byte buffer[256];
	byte* bloc = buffer;
	int lp1;
	int e;
	int ch;
#ifdef DEBUG
	printf("parseMem(%p,%d, %d)\n", mem, len, addrsz);
#endif
	
	/* Read address for data string */
	for (lp1=0; lp1<addrsz*2; lp1++) {
		ch = hexDigit(getChar());
		if (ch<0) return E_ADDRESSERROR;
		address = (address<<4) | ch;
		if (lp1&1) chk_upd(address & 0xFF);
	}
	
	/* Parse the data into buffer and get the result of the operation */
	e = parseData(buffer, len-addrsz);
#ifdef DEBUG
	printf("\treturn code %d from parsedata\n",e);
#endif
	if (e<0) return e; /* failed to correctly get data */
	
	/* if OK commit data to mem */
	mem = mem+address;
#ifdef DEBUG
	printf("\tCommitting %d bytes to address %x (real: %p)\n",len-addrsz,address,mem);
#endif
	for (; len>addrsz; len--) {
		*mem = *bloc;
		bloc++; mem++;
	}
	return E_SUCCESS;
}


int parseS5( byte* mem, int len ) {
	return E_NOSUPPORT;
}

int parseS7( byte* mem, int len ) {
	return E_NOSUPPORT;
}

int parseS8( byte* mem, int len ) {
	return E_NOSUPPORT;
}

int parseS9( byte* mem, int len ) {
	return E_NOSUPPORT;
}


int main() {
	byte memory[0x10000];
	
	printf("Starting...\n");
	parseRecs(memory);
	printf("Done parsing S-Records.\n");
	
	printf("SECRET_MESSAGE: '%s'\n", memory+0x2008 );
	
	return 0;
}
